import binascii
import hashlib

from communal.encoding import safe_encode


def crc32_unsigned(value):
    """binascii's crc32 is signed, make it unsigned"""
    return binascii.crc32(safe_encode(value)) & 0xFFFFFFFF


def checksum(filename, hashfunc="md5"):
    if not isinstance(hashfunc, str) or not hasattr(hashlib, hashfunc):
        raise ValueError("hashfunc must be a function specified in hashlib")

    hf = getattr(hashlib, hashfunc)
    h = hf()

    with open(filename, "rb") as f:
        while True:
            chunk = f.read(4096)
            if not chunk:
                break
            h.update(chunk)
    return h.hexdigest()



import os
import shutil

import fsspec
import furl
import requests

from communal.files import ensure_dir


def save_response_content(response, destination):
    with fsspec.open(destination, "wb") as f:
        shutil.copyfileobj(response.raw, f)


def download_file(url, dest, retries=3, retry_delay=5, session=requests, headers=None):
    url = furl.furl(url)

    exists = os.path.exists(dest)
    is_dir = os.path.isdir(dest) if exists else len(os.path.splitext(dest)[1]) == 0

    base_dir = dest if is_dir else os.path.dirname(dest)
    if base_dir:
        ensure_dir(base_dir)

    dest_filename = os.path.basename(dest) if not is_dir else None
    if not dest_filename and url.path.isfile:
        dest_filename = url.path.segments[-1]
        dest = os.path.join(dest, dest_filename)

    with session.get(url, headers=headers, stream=True) as resp:
        resp.raw.decode_content = True
        resp.raise_for_status()
        save_response_content(resp, dest)

    return dest



import socket

ip_address = None
try:
    from ipaddress import IPv4Address as ip_address
except ImportError:
    try:
        from ipaddr import IPAddress as ip_address
    except ImportError:
        pass


def current_host():
    return socket.gethostname()


def get_ip_for_host(host, ip_type=ip_address):
    return ip_type(socket.gethostbyname(host))


def current_ip_address(as_type=str):
    return get_ip_for_host(current_host(), ip_type=as_type)


def get_user_ip(request):
    remote_addr = request.remote_addr

    if "X-Original-Forwarded-For" in request.headers:
        remote_addr = request.headers.getlist("X-Original-Forwarded-For")[0].rpartition(
            " "
        )[-1]
    elif "X-Forwarded-For" in request.headers:
        remote_addr = request.headers.getlist("X-Forwarded-For")[0].rpartition(" ")[-1]

    if remote_addr:
        return ip_address(remote_addr)
    return remote_addr


def cidr_block_to_ipv4_range(network):
    addr, cidr = network.split("/")
    # Split address into octets and turn CIDR into int
    addr = addr.split(".")
    cidr = int(cidr)

    # Initialize the netmask and calculate based on CIDR mask
    mask = [0, 0, 0, 0]
    for i in range(cidr):
        mask[i // 8] = mask[i // 8] + (1 << (7 - i % 8))

    # Initialize net and binary and netmask with addr to get network
    net = []
    for i in range(4):
        net.append(int(addr[i]) & mask[i])

    broad = list(net)
    brange = 32 - cidr
    for i in range(brange):
        broad[3 - i // 8] = broad[3 - i // 8] + (1 << (i % 8))

    network_start = net[0] * 2**24 + net[1] * 2**16 + net[2] * 2**8 + net[3]
    network_end = broad[0] * 2**24 + broad[1] * 2**16 + broad[2] * 2**8 + broad[3]

    return (network_start, network_end)



import operator

from communal.collections import is_mapping
from communal.iterables import flatten


def key_getter(*attrs):
    attrs = [getattr(a, "key", a) for a in flatten(attrs)]
    itemgetter = operator.itemgetter(*attrs)
    attrgetter = operator.attrgetter(*attrs)

    def getter(obj):
        if not is_mapping(obj):
            return attrgetter(obj)
        else:
            return itemgetter(obj)

    return getter



from time import time

from communal.nulls import DoesNotExist


class classproperty(object):
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class cached_classproperty(classproperty):
    def __init__(self, getter):
        super().__init__(getter)
        self.cache = {}

    def __get__(self, instance, owner):
        if owner not in self.cache:
            self.cache[owner] = self.getter(owner)
        return self.cache[owner]


class cached_property(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Setting the ttl to a number expresses how long
    the property will last before being timed out (meaning it will be recomputed.)
    """

    def __init__(self, ttl=None, keep_cached_on_error=True):
        if callable(ttl):
            func = ttl
            ttl = None
        else:
            func = None
        self.ttl = ttl
        self.keep_cached_on_error = keep_cached_on_error
        self._prepare_func(func)

    def __call__(self, func):
        self._prepare_func(func)
        return self

    def __get__(self, obj, cls):
        if obj is None:
            return self

        now = time()
        obj_dict = obj.__dict__
        name = self.__name__
        ttl_expired = False
        value = DoesNotExist

        try:
            value, last_updated = obj_dict[name]
        except KeyError:
            pass
        else:
            ttl_expired = self.ttl and self.ttl < now - last_updated
            if not ttl_expired:
                return value

        try:
            value = self.func(obj)
        except Exception:
            if self.keep_cached_on_error and ttl_expired and value is not DoesNotExist:
                return value
            else:
                raise
        obj_dict[name] = (value, now)
        return value

    def __delete__(self, obj):
        obj.__dict__.pop(self.__name__, None)

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = (value, time())

    def _prepare_func(self, func):
        self.func = func
        if func:
            self.__doc__ = func.__doc__
            self.__name__ = func.__name__
            self.__module__ = func.__module__



from datetime import date, datetime

from dateutil.parser import parse as _dateutil_parse


def current_timestamp():
    return datetime.utcnow().timestamp()


def current_year():
    return date.today().year


parse_datetime = _dateutil_parse


def parse_date(dt):
    return _dateutil_parse(dt).date()


def parse_datetime_ignore_tz_microseconds(d):
    return _dateutil_parse(d, ignoretz=True)


def parse_datetime_ignore_tz_seconds(d):
    return _dateutil_parse(d, ignoretz=True).replace(microsecond=0)



import functools
import inspect
import itertools
import operator
import re
from collections import deque

from communal.exceptions import DecoratorIncompatibilityError
from communal.iterables import iterify
from communal.nulls import DoesNotExist


def nop(obj):
    return obj


def equal_to(other):
    return functools.partial(operator.eq, other)


def not_equal_to(other):
    return functools.partial(operator.ne, other)


def argmin(a):
    return min(range(len(a)), key=a.__getitem__)


def argmax(a):
    return max(range(len(a)), key=a.__getitem__)


is_none = functools.partial(operator.is_, None)
not_none = functools.partial(operator.is_not, None)


def if_none(default):
    def value(v):
        return v if v is not None else default

    return value


def stop_on_input_values(*vals, default=None):
    if not vals:
        vals = (None,)

    def deco(f):
        @functools.wraps(f)
        def func(v):
            return f(v) if v not in vals else default

        return func

    if len(vals) == 1 and callable(vals[0]):
        f = vals[0]
        vals = (None,)
        return deco(f)
    return deco


call_if_not_null = stop_on_input_values(None)


def stop_on_conditions(*conditions, default=None):
    def deco(f):
        @functools.wraps(f)
        def func(v):
            all_conditions_met = all((c(v) for c in conditions))
            return f(v) if all_conditions_met else default

        return func

    return deco


true_values = {"TRUE", "T", "Y", "YES", "ON", "1"}
false_values = {"FALSE", "F", "N", "NO", "OFF", "0"}


def str_to_bool(value: str):
    value = (value or "").upper()
    if value in true_values:
        return True
    elif value in false_values:
        return False
    else:
        return None


def attr_get(*keys: str, default=None):
    def get_attrs(o):
        return [getattr(o, k, default) for k in keys]

    if len(keys) == 1:
        key = keys[0]

    def get_attr(o):
        return getattr(o, key, default)

    return get_attrs if len(keys) > 1 else get_attr


def item_get(*keys: str, default=None):
    def get_keys(d):
        return [d.get(k, default) for k in keys]

    if len(keys) == 1:
        key = keys[0]

    def get_key(d):
        return d.get(key, default)

    return get_keys if len(keys) > 1 else get_key


def attr_or_key_getter(key, default=DoesNotExist):
    def with_obj(obj):
        val = getattr(obj, key, DoesNotExist)
        if val is DoesNotExist:
            try:
                val = obj[key]
                obj = val
            except (IndexError, TypeError):
                val = default
        else:
            obj = val
        return val

    return with_obj


def has_all_attrs(*attrs: str):
    def attrs_exist(o):
        return all((hasattr(o, a) for a in attrs))

    if len(attrs) == 1:
        attr = attrs[0]

    def attr_exists(o):
        return hasattr(o, attr)

    return attrs_exist if len(attrs) > 1 else attr_exists


def all_attrs_not_none(*attrs: str):
    def attrs_not_none(o):
        return all((getattr(o, a, None) is not None for a in attrs))

    if len(attrs) == 1:
        attr = attrs[0]

    def attr_not_none(o):
        return getattr(o, attr, None) is not None

    return attrs_not_none if len(attrs) > 1 else attr_not_none


def has_all_keys(*keys: str):
    def keys_exist(d):
        return all((k in d for k in keys))

    if len(keys) == 1:
        key = keys[0]

    def key_exists(d):
        return key in d

    return keys_exist if len(keys) > 1 else key_exists


def all_keys_not_none(*keys: str):
    def keys_not_none(d):
        return all((d.get(k, None) is not None for k in keys))

    if len(keys) == 1:
        key = keys[0]

    def key_not_none(d):
        return d.get(key, None) is not None

    return keys_not_none if len(keys) > 1 else key_not_none


def map_values(d: dict):
    def mapped_value(val):
        return d.get(val, val)

    return mapped_value


# Because partials are only specifying the first argument to a function,
# and we're effectively specifying the righthand side up front and then
# running the function on a bunch of inputs, we use the inverse functions
# e.g. less_than(5) will run x < 5 for any x. To make the arguments left-to-right,
# we end up calling the equivalent operator.gt(5, x), or is 5 > x


def less_than(other):
    return functools.partial(operator.gt, other)


def greater_than(other):
    return functools.partial(operator.lt, other)


def less_than_or_equal_to(other):
    return functools.partial(operator.ge, other)


def greater_than_or_equal_to(other):
    return functools.partial(operator.le, other)


def regex_sub(pattern, repl):
    pattern = re.compile(pattern)
    return functools.partial(pattern.sub, repl)


class Pipeline(object):
    def __init__(self, *functions):
        self.functions = functions

    def __repr__(self):
        return "Pipeline: [{}]".format(
            ", ".join((str(func) for func in self.functions))
        )

    def __call__(self, val):
        for f in self.functions:
            val = f(val)
        return val


class FunctionMap(object):
    def __init__(self, *functions, use_tuple=True):
        self.functions = functions
        self.use_tuple = use_tuple

    def __repr__(self):
        return f'{self.__class__.__name__}: [{", ".join((str(func) for func in self.functions))}]'

    def __call__(self, *args, **kw):
        val = [func(*args, **kw) for func in self.functions]
        if self.use_tuple:
            return tuple(val)
        else:
            return val


class FunctionZipMap(FunctionMap):
    def __call__(self, vals, **kw):
        val = [func(arg, **kw) for func, arg in zip(self.functions, vals)]

        if self.use_tuple:
            return tuple(val)
        else:
            return val


class FunctionStarMap(object):
    def __init__(self, function, use_tuple=True):
        self.function = function
        self.use_tuple = use_tuple

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.function}"

    def __call__(self, *args, **kw):
        val = list(
            itertools.chain(*itertools.starmap(self.function, itertools.chain(*args)))
        )
        if self.use_tuple:
            return tuple(val)
        else:
            return val

        return ()


class VectorFunction(object):
    def __init__(self, function):
        self.function = function

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.function}"

    def __call__(self, values):
        return list(map(self.function, values))


class ChainedFunction(object):
    def __init__(self, function):
        self.function = function

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.function}"

    def __call__(self, values):
        return list(itertools.chain(*(map(self.function, iterify(v)) for v in values)))


class hybridmethod(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        self_or_cls = owner if instance is None else instance

        @functools.wraps(self.func)
        def call_hybrid(*args, **kwargs):
            return self.func(self_or_cls, *args, **kwargs)

        return call_hybrid


def get_innermost_argspec(method):
    """When passed a method/function, drill through all layers of decorators to get the inner argspec."""

    orig_method = method
    q = deque([orig_method])
    while q:
        method = q.popleft()
        if hasattr(method, "__func__"):
            method = method.__func__
        elif hasattr(method, "func"):
            method = method.func

        argspec = inspect.getfullargspec(method)
        args = argspec.args

        if args and args[0] in ("self", "cls"):
            return argspec
        if not hasattr(method, "__closure__") or method.__closure__ is None:
            raise DecoratorIncompatibilityError(
                "Decorator has no closure: {}".format(method)
            )

        closure = method.__closure__

        for cell in closure:
            inner_method = cell.cell_contents
            if inner_method is method:
                continue
            if not inspect.isfunction(inner_method) and not inspect.ismethod(
                inner_method
            ):
                continue
            q.append(inner_method)



from collections import deque
from collections.abc import Generator, Iterable, Iterator, Sequence
from itertools import (
    chain,
    combinations,
    count,
    cycle,
    filterfalse,
    islice,
    tee,
)
from typing import List


def is_sequence(obj):
    return isinstance(obj, Sequence) and not isinstance(obj, (str, bytes))


def is_iterable(obj):
    return isinstance(obj, Iterable) and not isinstance(obj, (str, bytes))


def iterify(o, empty_list_if_none=True):
    if o is None and empty_list_if_none:
        return []
    return [o] if not is_iterable(o) else o


def listify(o, empty_list_if_none=True):
    if o is None and empty_list_if_none:
        return []
    if isinstance(o, (range, Generator, Iterator)):
        return list(o)

    return [o] if not is_sequence(o) else o


def flatten(obj):
    return chain.from_iterable(iterify(obj))


def flatten_args(*args):
    result = []

    for arg in args:
        if is_sequence(arg):
            result.extend(arg)
        else:
            result.append(arg)

    return result


def combine_lists(*args: List):
    return sum(args, [])


def pairwise(iterable):
    """
    s -> (s0,s1), (s1,s2), (s2, s3), ...
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def partition(pred, iterable):
    """
    Use a predicate to partition entries into false entries and true entries
    """
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)


def powerset(iterable):
    """
    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def round_robin(*iterables):
    """
    roundrobin('ABC', 'D', 'EF') --> A D E B F C
    """
    # Recipe credited to George Sakkis
    pending = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while pending:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))


def all_combinations(s):
    result = [tuple(s)]

    q = deque([([], s)])

    while q:
        left, right = q.popleft()

        res = [
            (i, i + k)
            for k in range(2, len(right) + 1)
            for i in range(len(right) - k + 1)
        ]

        for i, j in res:
            new_left = tuple(left) + tuple(right[:i]) + (tuple(right[i:j]),)
            new_right = tuple(right[j:])
            result.append(new_left + new_right)
            q.append((new_left, new_right))

    return result


def batch_iter(iterable, batch_size: int):
    source_iter = iter(iterable)
    while True:
        batch = list(islice(source_iter, batch_size))
        if len(batch) > 0:
            yield batch
        else:
            return


def iter_len(it):
    counter = count()
    deque(zip(it, counter), maxlen=0)
    return next(counter)


def argmax(it):
    return max(range(len(it)), key=it.__getitem__)


def split_index_on_condition(it, condition):
    condition_true = []
    condition_false = []

    for i, r in enumerate(it):
        if condition(r):
            condition_true.append(i)
        else:
            condition_false.append(i)

    return condition_true, condition_false


def split_on_condition(it, condition):
    condition_true = []
    condition_false = []

    for r in it:
        if condition(r):
            condition_true.append(r)
        else:
            condition_false.append(r)

    return condition_true, condition_false


def split_on_indices(it, indices, sort_indices=False):
    if sort_indices:
        indices = sorted(indices)
    elif not is_sequence(indices):
        indices = listify(indices)

    m = len(indices)
    data = listify(it)
    if m == 0:
        return [], data
    elif m == len(data):
        return data, []

    j = 0
    ind = indices[j]

    match = []
    no_match = []

    for i, x in enumerate(data):
        if i < ind:
            no_match.append(x)
        elif i == ind:
            j += 1
            match.append(x)
            if j >= m:
                match.extend(data[i + 1 :])
                break
            ind = indices[j]
    return match, no_match


def merge_values_at_indices(it, indices, values, sort_indices=False):
    if sort_indices:
        indices = sorted(indices)
    elif not is_sequence(indices):
        indices = listify(indices)

    m = len(indices)
    other = listify(values)
    if len(other) != m:
        raise ValueError(
            f"Length of merged sequence ({len(values)}) must be equal to length of indices {m}"
        )

    if m == 0:
        return listify(it)

    result = []

    if is_sequence(it):
        it = iter(it)

    for index, value in zip(indices, values):
        while len(result) < index:
            result.append(next(it))
        result.append(value)
    return result



def safe_decode(
    value, encoding="utf-8", errors="strict", text_types=str, binary_types=bytes
):
    if isinstance(value, text_types):
        return value

    if isinstance(value, binary_types):
        return value.decode(encoding, errors)
    else:
        return str(value)


def safe_encode(
    value,
    encoding="utf-8",
    from_encoding=None,
    errors="strict",
    text_types=str,
    binary_types=bytes,
):
    if isinstance(value, text_types):
        return value.encode(encoding, errors=errors)
    elif isinstance(value, binary_types):
        if hasattr(from_encoding, "lower"):
            from_encoding = from_encoding.lower()
        if hasattr(encoding, "lower"):
            encoding = encoding.lower()

        if value and from_encoding and encoding and encoding != from_encoding:
            value = safe_decode(value, encoding=from_encoding, errors=errors)
            return value.encode(encoding, errors)
        return value
    else:
        return bytes(repr(value), encoding=encoding)



import importlib
import pkgutil


def import_submodules(package, recursive=True):
    """Import all submodules of a module, recursively, including subpackages

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    q = [package]

    results = {}

    while q:
        package = q.pop()

        if isinstance(package, str):
            package = importlib.import_module(package)

        for _loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
            full_name = package.__name__ + "." + name
            results[full_name] = importlib.import_module(full_name)
            if recursive and is_pkg:
                q.append(full_name)
    return results



import re

camel_to_snake_re_all_caps = re.compile("([A-Z\\d]+)([A-Z][a-z])")
camel_to_snake_re_before_word = re.compile("([a-z])([A-Z])")


def camel_to_snake_case(name):
    name = camel_to_snake_re_all_caps.sub("\\1_\\2", name)
    name = camel_to_snake_re_before_word.sub("\\1_\\2", name)
    name = name.replace("-", "_")
    return name.lower()


def to_camel_case(name):
    return name.title().replace(" ", "").replace("_", "")


def to_title_case(name):
    """Convert all upper or all lower case names to title case, otherwise preserve casing."""
    if name.islower() or name.isupper():
        return name.title()
    else:
        return name



def update_from_bases(
    cls,
    name,
    bases,
    initial_cls=dict,
    override_with_child=False,
    update_with_child=True,
):
    if name in cls.__dict__ and override_with_child:
        return cls.__dict__[name]

    val = initial_cls()
    for base in bases:
        if base.__dict__.get(name):
            val.update(base.__dict__[name])

    if update_with_child and cls.__dict__.get(name):
        val.update(cls.__dict__[name])

    return val



class _Omitted:
    def __bool__(self):
        return False

    def __repr__(self):
        return "Omitted"


_Omitted.__nonzero__ = _Omitted.__bool__

Omitted = _Omitted()


class _DoesNotExist:
    def __bool__(self):
        return False

    def __repr__(self):
        return "DoesNotExist"


_DoesNotExist.__nonzero__ = _DoesNotExist.__bool__

DoesNotExist = _DoesNotExist()


class NotInSet:
    pass


def obj_or_none(obj):
    return obj or None



import os
import zipfile

import fsspec

from communal.iterables import iterify


def unzip_all_files(zip_filename, out_dir=None, full_path=True):
    if out_dir is None:
        out_dir = os.path.dirname(zip_filename) or os.curdir

    if os.path.isdir(out_dir) and not out_dir.endswith(os.sep):
        out_dir += os.sep

    paths = []

    with fsspec.open(zip_filename).open() as f:
        zf = zipfile.ZipFile(f)

        for path in zf.infolist():
            filename = path.filename
            base_filename = os.path.basename(filename)
            extract_path = filename if full_path else base_filename
            out_filename = os.path.join(out_dir, extract_path)
            zf.extract(extract_path, out_dir)
            paths.append(out_filename)
    return paths


def extract_filenames_from_zip(zip_filename, filenames, out_dir, full_path=True):
    filenames = iterify(filenames)

    if os.path.isdir(out_dir) and not out_dir.endswith(os.sep):
        out_dir += os.sep

    with fsspec.open(zip_filename).open() as f:
        zf = zipfile.ZipFile(f)

        paths = []

        for path in zf.infolist():
            filename = path.filename
            base_filename = os.path.basename(filename)
            if filename in filenames or base_filename in filenames:
                extract_path = filename if full_path else base_filename
                out_filename = os.path.join(out_dir, extract_path)
                zf.extract(extract_path, out_dir)
                paths.append(out_filename)

    if len(filenames) > 1:
        return paths
    elif len(filenames) > 0:
        return paths[0]
    else:
        return None



from collections import OrderedDict
from collections.abc import Hashable, Mapping


def unique_list(values):
    """Produces a unique list while preserving input order"""
    return list(OrderedDict.fromkeys(values))


def is_mapping(val):
    return isinstance(val, Mapping)


def is_hashable(val):
    return isinstance(val, Hashable)



import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry


def add_retries_to_requests_session(
    session, retries=3, backoff_factor=0.3, status_forcelist=(502, 503, 504)
):
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)


def requests_session_with_retries(
    retries=3, backoff_factor=0.3, status_forcelist=(502, 503, 504)
):
    session = requests.Session()
    add_retries_to_requests_session(
        session,
        retries=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    return session



import inspect


class MixinHooks:
    __include_subclass__ = True

    @classmethod
    def call_mixin_hooks(cls, bases=None, **kwargs):
        for base in bases or inspect.getmro(cls):
            if "mixin_hook" in base.__dict__:
                base.mixin_hook(cls, **kwargs)

    @classmethod
    def all_base_classes(cls, with_cls=None):
        if with_cls is None:
            with_cls = cls.__include_subclass__

        mro = inspect.getmro(cls)
        return mro if with_cls else mro[1:]



import enum


class StringEnum(str, enum.Enum):
    def __str__(self):
        return str(self.value)

    @classmethod
    def display_name(cls, choice):
        return choice.name

    @property
    def display(self):
        return self.value.replace("_", " ").title()

    @classmethod
    def choices(cls):
        return [(choice.value, cls.display_name(choice)) for choice in cls]

    @classmethod
    def from_value(cls, value):
        return cls(value)


class LabeledEnum(enum.Enum):
    label: str

    def __new__(cls, value, label):
        obj = object.__new__(cls)
        obj._value = value
        obj.label = label
        return obj



from datetime import datetime

import pendulum
from pytz import timezone

US_EASTERN = timezone("US/Eastern").zone
US_CENTRAL = timezone("US/Central").zone
US_MOUNTAIN = timezone("US/Mountain").zone
US_PACIFIC = timezone("US/Pacific").zone
UTC = timezone("UTC").zone


def local_datetime_utc(*args, **kwargs):
    dt = pendulum(*args, **kwargs)
    return datetime.utcfromtimestamp(dt.timestamp())



import os
import shutil

import fsspec

from communal.sizes import COPY_BUFFER_SIZE


def ensure_dir(d):
    if not os.path.exists(d):
        os.makedirs(d)


def copy_file(filename, dest, buffer_size=COPY_BUFFER_SIZE):
    dest_kwargs = {}
    if dest.endswith(".gz") and not filename.endswith(".gz"):
        dest_kwargs.update(compression="gzip")
    elif dest.endswith(".zip") and not filename.endswith(".zip"):
        dest_kwargs.update(compression="zip")

    with fsspec.open(dest, "wb", **dest_kwargs) as df:
        with fsspec.open(filename, "rb") as sf:
            shutil.copyfileobj(sf, df)
    return dest


def remove_file(filename):
    os.unlink(filename)


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)



class AttrDict(dict):
    """
    A dictionary that allows using the dot operator to get and set keys.
    """

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __repr__(self):
        return f"{self.__class__.__name__}({dict.__repr__(self)})"



class ClientProxy(object):
    client = None

    def __getattr__(self, name):
        return getattr(self.client, name)

    def __getitem__(self, name):
        return self.client[name]

    def __setitem__(self, name, value):
        self.client[name] = value

    def __delitem__(self, name):
        del self.client[name]



import copy
import operator
from collections import deque

from communal.collections import is_mapping
from communal.functions import attr_or_key_getter
from communal.iterables import is_sequence, iterify
from communal.nulls import DoesNotExist, Omitted


def deep_merge(d, *others, overwrite=True):
    result = copy.deepcopy(d)

    queue = deque([(d_i, ()) for d_i in others])
    while queue:
        d, path = queue.popleft()
        d_res = nested_get(result, path)
        if not d_res:
            d_res = {}
            nested_set(result, path, d_res)

        for k, v in d.items():
            if is_mapping(v):
                queue.append((v, path + (k,)))
            elif overwrite or k not in d_res:
                d_res[k] = v

    return result


def nested_diff(orig, other):
    result = {}

    queue = deque([(other, ())])
    while queue:
        d, path = queue.popleft()
        d_res = nested_get(orig, path)
        if not d_res:
            d_res = {}
            nested_set(result, path, d_res)

        for k, v in d.items():
            if is_mapping(v):
                queue.append((v, path + (k,)))
            elif k not in d_res:
                d_res[k] = v

    return result


def nested_getattr(obj, keys, default=None):
    use_map = False

    if isinstance(keys, str):
        keys = keys.split(".")

    for key in keys:
        is_string = isinstance(key, str)
        is_digit = is_string and key.isdigit()
        is_star = key == "*"

        if is_string and not is_digit and not is_star:
            func = attr_or_key_getter(key)
            if not use_map:
                val = func(obj)
                obj = val
            else:
                val = list(map(func, obj))
                obj = val
        elif is_sequence(obj) and (isinstance(key, int) or is_digit):
            try:
                val = obj[int(key)]
                obj = val
            except (IndexError, TypeError):
                obj = DoesNotExist
        elif is_sequence(obj) and is_star:
            use_map = True
        else:
            raise ValueError("Paths must be string attributes or int indexes")
        if obj is DoesNotExist or obj is None:
            return default
    return obj


def nested_get(obj, keys, default=DoesNotExist):
    if isinstance(keys, str):
        keys = keys.split(".")

    # For source_data fields, just return the whole document
    if len(keys) == 0:
        return obj

    use_map = False

    try:
        for key in keys:
            is_string = isinstance(key, str)
            is_digit = is_string and key.isdigit()
            is_star = key == "*"

            if is_string and not is_digit and not is_star:
                if is_mapping(obj) and not use_map:
                    val = obj.get(key, default)
                    obj = val
                elif is_sequence(obj) and use_map:
                    val = list(map(operator.itemgetter(key), obj))
                    obj = val
            elif is_sequence(obj) and (
                isinstance(key, int) or (isinstance(key, str) and key.isdigit())
            ):
                try:
                    val = obj[int(key)]
                    obj = val
                except IndexError:
                    return default
            elif is_sequence(obj) and is_star:
                use_map = True
            else:
                return default
        if obj is DoesNotExist or obj is None:
            return default
        return obj
    except AttributeError:
        return default


def nested_getter(default=DoesNotExist):
    base_default = default

    def _nested_get_with_default(obj, keys, default=Omitted):
        default = default if default is not Omitted else base_default
        return nested_get(obj, keys, default=default)

    return _nested_get_with_default


def nested_exists(d, key):
    keys = iterify(key)
    obj = nested_get(d, keys[:-1], default=DoesNotExist)

    last_key = keys[-1]

    if is_mapping(obj):
        return last_key in obj
    elif is_sequence(obj) and (
        isinstance(last_key, int) or (isinstance(last_key, str) and last_key.isdigit())
    ):
        return len(obj) > int(last_key)

    return False


def nested_set(d, keys, value):
    level = d

    key_pairs = list(zip(keys, keys[1:]))
    for key, next_key in key_pairs:
        key_is_string = isinstance(key, str)
        key_is_digit = isinstance(key, int) or (key_is_string and key.isdigit())

        next_key_is_string = isinstance(next_key, str)
        next_key_is_digit = isinstance(next_key, int) or (
            next_key_is_string and next_key.isdigit()
        )

        next_level = None

        if key_is_string:
            next_level = level.get(key, None)
            if next_level is None:
                if next_key_is_string:
                    level[key] = {}
                    next_level = level[key]
                elif next_key_is_digit:
                    next_key = int(next_key)
                    level[key] = [None] * next_key
                next_level = level[key]
        elif key_is_digit:
            key = int(key)
            if len(level) > key:
                next_level = level[key]
            else:
                if len(level) > 0:
                    level.extend([None] * (key + 1 - len(level)))
                else:
                    level.extend([None] * (key + 1))

            if next_level is None:
                if next_key_is_string:
                    level[key] = {}
                    next_level = level[key]
                elif next_key_is_digit:
                    next_key = int(next_key)
                    level[key] = [None] * next_key
                next_level = level[key]

        level = next_level

    last_key = keys[-1]
    if isinstance(last_key, int) or (isinstance(last_key, str) and last_key.isdigit()):
        last_key = int(last_key)
        if len(level) <= last_key:
            if len(level) > 0:
                level.extend([None] * (last_key + 1 - len(level)))
            else:
                level.extend([None] * (last_key + 1))
    level[last_key] = value


def nested_setattr(obj, keys, value):
    if len(keys) > 0:
        obj = nested_getattr(obj, keys[:-1])
        last_key = keys[-1]
        is_string = isinstance(last_key, str)
        is_digit = isinstance(last_key, int) or (is_string and last_key.isdigit())
        if is_string and not is_digit:
            setattr(obj, last_key, value)
        elif isinstance(last_key, int) or (is_string and is_digit):
            obj[int(last_key)] = value


def nested_del(d, keys):
    level = d
    for key in keys[:-1]:
        next_level = level.get(key, None)
        if next_level is None:
            return
        level = next_level
    del level[keys[-1]]


def nested_delattr(obj, keys):
    if len(keys) > 0:
        for key in keys[:-1]:
            obj = getattr(obj, key)
        delattr(obj, keys[-1])


class Path(object):
    def __init__(self, path, separator="."):
        if isinstance(path, Path):
            self.path = path.path
        else:
            self.path = path
        self.separator = separator

    @property
    def parts(self):
        return self.path.split(self.separator)

    def __iter__(self):
        for part in self.parts:
            yield part

    def __len__(self):
        return len(self.parts)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self.path)

    def index(self, element):
        return self.parts.index(element)

    def __getitem__(self, slice):
        result = self.parts[slice]
        if isinstance(result, list):
            return self.__class__(self.separator.join(result), separator=self.separator)
        return result

    def __eq__(self, other):
        return self.path == other.path and self.separator == other.separator

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return self.path



import resource
import sys


def get_current_process_memory_usage():
    """Gets current process memory usage (residence set size) in megabytes."""
    resource_data = resource.getrusage(resource.RUSAGE_SELF)
    if sys.platform == "darwin":
        return resource_data.ru_maxrss / (1024 * 1024)
    else:
        return resource_data.ru_maxrss / 1024



import base64

from communal.encoding import safe_decode, safe_encode
from communal.functions import Pipeline
from communal.properties import classproperty


def int_to_bytes(x: int, endianness="little"):
    if not isinstance(x, int):
        x = int(x)
    return x.to_bytes((x.bit_length() + 7) // 8, endianness)


def int_from_bytes(x: bytes, endianness="little"):
    return int.from_bytes(x, endianness)


class BinaryEncoder(object):
    @classmethod
    def strip_padding(cls, s):
        return s.rstrip(b"=")

    @classproperty
    def encode(cls):
        return Pipeline(cls.encoder, cls.strip_padding)

    @classmethod
    def add_padding(cls, s):
        return s if len(s) % cls.group_width == 0 else s.ljust(cls.group_width, b"=")

    @classproperty
    def decode(cls):
        return Pipeline(cls.add_padding, cls.decoder)

    @classmethod
    def encode_str(cls, s: str):
        return safe_decode(cls.encode(safe_encode(s)))

    @classmethod
    def encode_int(cls, x: int, endianness: str = "little"):
        return safe_decode(cls.encode(int_to_bytes(x, endianness=endianness)))

    @classmethod
    def decode_str(cls, s: str):
        return safe_decode(cls.decode(safe_encode(s)))

    @classmethod
    def decode_int(cls, s: str, endianness: str = "little"):
        return int_from_bytes(cls.decode(safe_encode(s)), endianness=endianness)


class Base16(BinaryEncoder):
    bits = 16
    group_width = 2
    encoder = base64.b16encode
    decoder = base64.b16decode


class Base32(BinaryEncoder):
    bits = 32
    group_width = 8
    encoder = base64.b32encode
    decoder = base64.b32decode


class Base64(BinaryEncoder):
    bits = 64
    group_width = 8
    encoder = base64.b64encode
    decoder = base64.b64decode


class Base85(BinaryEncoder):
    bits = 85
    encoder = base64.b85encode
    decoder = base64.b85decode

    encode = encoder
    decode = decoder
